#include "viewer.hpp"
#include <iostream>
#include <GL/gl.h>
#include <GL/glu.h>
#include "draw.hpp"

Viewer::Viewer(): point1(), point2() {
  Glib::RefPtr<Gdk::GL::Config> glconfig;

  // Ask for an OpenGL Setup with
  //  - red, green and blue component colour
  //  - a depth buffer to avoid things overlapping wrongly
  //  - double-buffered rendering to avoid tearing/flickering
  glconfig = Gdk::GL::Config::create(Gdk::GL::MODE_RGB |
                                     Gdk::GL::MODE_DEPTH |
                                     Gdk::GL::MODE_DOUBLE);
  if (glconfig.is_null()) {
    // If we can't get this configuration, die
    std::cerr << "Unable to setup OpenGL Configuration!" << std::endl;
    abort();
  }

  // Accept the configuration
  set_gl_capability(glconfig);

  // Register the fact that we want to receive these events
  add_events(Gdk::BUTTON1_MOTION_MASK    |
             Gdk::BUTTON2_MOTION_MASK    |
             Gdk::BUTTON3_MOTION_MASK    |
             Gdk::BUTTON_PRESS_MASK      | 
             Gdk::BUTTON_RELEASE_MASK    |
             Gdk::VISIBILITY_NOTIFY_MASK);
  cube_vertex[0] = Point3D( 1.0, 1.0, 1.0);
  cube_vertex[1] = Point3D(-1.0, 1.0, 1.0);
  cube_vertex[2] = Point3D(-1.0, 1.0,-1.0);
  cube_vertex[3] = Point3D(-1.0,-1.0,-1.0);
  cube_vertex[4] = Point3D(-1.0,-1.0, 1.0);
  cube_vertex[5] = Point3D( 1.0,-1.0, 1.0);
  cube_vertex[6] = Point3D( 1.0,-1.0,-1.0);
  cube_vertex[7] = Point3D( 1.0, 1.0,-1.0);
  
  button_down[0] = false;
  button_down[1] = false;
  button_down[2] = false;

  Glib::signal_timeout().connect(SigC::slot(*this, &Viewer::invalidate), 20);
  
  reset_view_no_invalidate();
}

Viewer::~Viewer() {
  // Nothing to do here right now.
}

bool Viewer::invalidate() {
  // Force a rerender
  GtkAllocation allocation = get_allocation();
  get_window()->invalidate_rect(Gdk::Rectangle(&allocation), false);
  return true;
}

void Viewer::set_perspective(double fov, double aspect,
                             double near, double far) {
  double radians = fov*M_PI/180;
  m_projection[0][0] = 1/tan(radians/2)/aspect;
  m_projection[1][1] = 1/tan(radians/2);
  m_projection[2][2] = -(far+near)/(far-near);
  m_projection[2][3] = -2*far*near/(far-near);
  m_projection[3][2] = -1;
  m_projection[3][3] = 0;
}

void Viewer::reset_view() {
  reset_view_no_invalidate();
  //invalidate();
}

void Viewer::reset_view_no_invalidate() {
  m_model = Matrix4x4();
  scale[0] = 1.0;
  scale[1] = 1.0;
  scale[2] = 1.0;
  Matrix4x4 reflect;
  reflect[2][2] = -1.0;
  m_view = Matrix4x4() * translation(Vector3D(0.0,0.0,10.0)) * reflect;
  viewport_coord[0][0] = 0.05 * get_width();
  viewport_coord[0][1] = 0.05 * get_height();
  viewport_coord[1][0] = 0.95 * get_width();
  viewport_coord[1][1] = 0.95 * get_height();
  near_clip = 5.0;
  far_clip = 15.0;
  max_near_clip = 0.1;
  max_far_clip = 20.0;
  fov = 30.0;
  set_perspective(fov, get_view_width()/get_view_height(),near_clip,far_clip);
}

void Viewer::on_realize() {
  // Do some OpenGL setup.
  // First, let the base class do whatever it needs to
  Gtk::GL::DrawingArea::on_realize();
  
  Glib::RefPtr<Gdk::GL::Drawable> gldrawable = get_gl_drawable();
  
  if (!gldrawable)
    return;

  if (!gldrawable->gl_begin(get_gl_context()))
    return;
  
  viewport_coord[0][0] = 0.05 * get_width();
  viewport_coord[0][1] = 0.05 * get_height();
  viewport_coord[1][0] = 0.95 * get_width();
  viewport_coord[1][1] = 0.95 * get_height();

  gldrawable->gl_end();
}

bool Viewer::on_expose_event(GdkEventExpose* event) {
  Glib::RefPtr<Gdk::GL::Drawable> gldrawable = get_gl_drawable();

  if (!gldrawable) return false;

  if (!gldrawable->gl_begin(get_gl_context()))
    return false;

  // Here is where your drawing code should go.
  
  draw_init(get_width(), get_height());
  
  /* A few of lines are drawn below to show how it's done. */

  
  draw_model_gnomon();
  draw_modified_cube();
  draw_view_gnomon();
  draw_viewport();

  draw_complete();
            
  // Swap the contents of the front and back buffers so we see what we
  // just drew. This should only be done if double buffering is enabled.
  gldrawable->swap_buffers();

  gldrawable->gl_end();

  return true;
}

void Viewer::draw_modified_cube() {
  
  Point3D modified_cube[8];
  for(int i = 0; i < 8; i++) {
    for(int j = 0; j < 3; j++) {
      modified_cube[i][j] = cube_vertex[i][j]*scale[j];
    }
    modified_cube[i] = m_view*m_model*modified_cube[i];
  }
  
  set_colour(Colour(1.0, 1.0, 1.0));
  for(int i = 0; i < 4; i++) {   
    my_draw_line(modified_cube[i*2], modified_cube[(i*2+7)%8]);
    my_draw_line(modified_cube[i*2], modified_cube[(i*2+1)%8]);
    my_draw_line(modified_cube[i*2], modified_cube[(i*2+5)%8]);
  }
}

void Viewer::draw_model_gnomon() {
  set_colour(Colour(1.0, 0.0, 0.0));
  my_draw_line(m_view*m_model*Point3D(0.0,0.0,0.0), m_view*m_model*Point3D(0.5,0.0,0.0));
  set_colour(Colour(0.0, 1.0, 0.0));
  my_draw_line(m_view*m_model*Point3D(0.0,0.0,0.0), m_view*m_model*Point3D(0.0,0.5,0.0));
  set_colour(Colour(0.0, 0.0, 1.0));
  my_draw_line(m_view*m_model*Point3D(0.0,0.0,0.0), m_view*m_model*Point3D(0.0,0.0,0.5));
}

void Viewer::draw_view_gnomon() {
  set_colour(Colour(1.0, 0.0, 0.0));
  my_draw_line(m_view*Point3D(0.0,0.0,0.0), m_view*Point3D(0.5,0.0,0.0));
  set_colour(Colour(0.0, 1.0, 0.0));
  my_draw_line(m_view*Point3D(0.0,0.0,0.0), m_view*Point3D(0.0,0.5,0.0));
  set_colour(Colour(0.0, 0.0, 1.0));
  my_draw_line(m_view*Point3D(0.0,0.0,0.0), m_view*Point3D(0.0,0.0,0.5));
}

void Viewer::my_draw_line(Point3D a, Point3D b) { 
  //clip front
  double wecA = (a-Point3D(0,0,near_clip)).dot(Vector3D(0.0,0.0,1.0));
  double wecB = (b-Point3D(0,0,near_clip)).dot(Vector3D(0.0,0.0,1.0));
  if(wecA < 0 && wecB < 0) {
    return;
  }
  if(wecA < 0) {
    double t = wecA/(wecA - wecB);
    a = a + t*(b-a);
  }
  if(wecB < 0) {
    double t = wecA/(wecA - wecB);
    b = a + t*(b-a);
  }
  
  //clip back
  wecA = (a-Point3D(0,0,far_clip)).dot(Vector3D(0.0,0.0,-1.0));
  wecB = (b-Point3D(0,0,far_clip)).dot(Vector3D(0.0,0.0,-1.0));
  if(wecA < 0 && wecB < 0) {
    return;
  }
  if(wecA < 0) {
    double t = wecA/(wecA - wecB);
    a = a + t*(b-a);
  }
  if(wecB < 0) {
    double t = wecA/(wecA - wecB);
    b = a + t*(b-a);
  }

  double z1 = a[2];
  double z2 = b[2];
  
  //apply fov
  set_perspective(fov, get_view_width()/get_view_height(),near_clip,far_clip);
  a=m_projection*a;
  b=m_projection*b;
    
  Point2D first;
  Point2D second; 

  first[0] = a[0]/z1;
  first[1] = a[1]/z1;
  second[0] = b[0]/z2;
  second[1] = b[1]/z2;
  
  if(first[0] < -1.0 && second[0] < -1.0) {
    return;
  }
  if(first[0] > 1.0 && second[0] > 1.0) {
    return;
  }
  if(first[1] < -1.0 && second[1] < -1.0) {
    return;
  }
  if(first[1] > 1.0 && second[1] > 1.0) {
    return;
  }
  
  if(first[0] > 1.0) {
    double porp = (1.0-second[0])/(first[0]-second[0]);
    first[0] = 1.0;
    first[1] = (first[1]-second[1])*porp+second[1];
  }
  if(second[0] > 1.0) {
    double porp = (1.0-first[0])/(second[0]-first[0]);
    second[0] = 1.0;
    second[1] = (second[1]-first[1])*porp+first[1];
  }
  
  if(first[1] > 1.0) {
    double porp = (1.0-second[1])/(first[1]-second[1]);
    first[1] = 1.0;
    first[0] = (first[0]-second[0])*porp+second[0];
  }
  if(second[1] > 1.0) {
    double porp = (1.0-first[1])/(second[1]-first[1]);
    second[1] = 1.0;
    second[0] = (second[0]-first[0])*porp+first[0];
  }
  
  if(first[0] < -1.0) {
    double porp = (-1.0-second[0])/(first[0]-second[0]);
    first[0] = -1.0;
    first[1] = (first[1]-second[1])*porp+second[1];
  }
  if(second[0] < -1.0) {
    double porp = (-1.0-first[0])/(second[0]-first[0]);
    second[0] = -1.0;
    second[1] = (second[1]-first[1])*porp+first[1];
  }
  
  if(first[1] < -1.0) {
    double porp = (-1.0-second[1])/(first[1]-second[1]);
    first[1] = -1.0;
    first[0] = (first[0]-second[0])*porp+second[0];
  }
  if(second[1] < -1.0) {
    double porp = (-1.0-first[1])/(second[1]-first[1]);
    second[1] = -1.0;
    second[0] = (second[0]-first[0])*porp+first[0];
  }
  
  first[0] = first[0]*get_view_width()/2+get_view_center()[0];
  first[1] = -first[1]*get_view_height()/2+get_view_center()[1];
  second[0] = second[0]*get_view_width()/2+get_view_center()[0];
  second[1] = -second[1]*get_view_height()/2+get_view_center()[1];
  
  draw_line(first, second);
}

double Viewer::get_view_width() {
  return viewport_coord[1][0] - viewport_coord[0][0];
}

double Viewer::get_view_height() {
  return viewport_coord[1][1] - viewport_coord[0][1];
}

void Viewer::draw_viewport() {
  set_colour(Colour(0.0, 0.0, 0.0));

  draw_line(viewport_coord[0],
            Point2D(viewport_coord[0][0], viewport_coord[1][1]));
  draw_line(viewport_coord[0],
            Point2D(viewport_coord[1][0], viewport_coord[0][1]));
  draw_line(viewport_coord[1],
            Point2D(viewport_coord[0][0], viewport_coord[1][1]));
  draw_line(viewport_coord[1],
            Point2D(viewport_coord[1][0], viewport_coord[0][1]));
}

bool Viewer::on_configure_event(GdkEventConfigure* event) {
  Glib::RefPtr<Gdk::GL::Drawable> gldrawable = get_gl_drawable();

  if (!gldrawable) return false;
  
  if (!gldrawable->gl_begin(get_gl_context()))
    return false;

  gldrawable->gl_end();

  return true;
}

bool Viewer::on_button_press_event(GdkEventButton* event) {
  //std::cerr << "Stub: Button " << event->button << " pressed" << std::endl;
  point1[0] = event->x;
  point1[1] = event->y;
  point2[0] = event->x;
  point2[1] = event->y;
  button_down[event->button - 1] = true;
  return true;
}

bool Viewer::on_button_release_event(GdkEventButton* event) {
  //std::cerr << "Stub: Button " << event->button << " released" << std::endl;
  if(mode == VIEWPORT && button_down[0]) {
    point2 = point1;
    point1[0] = event->x;
    point1[1] = event->y;
    viewport_coord[0][0] = my_min(point1[0],point2[0]);
    viewport_coord[0][0] = my_max(0.0,viewport_coord[0][0]);
  
    viewport_coord[0][1] = my_min(point1[1],point2[1]);
    viewport_coord[0][1] = my_max(0.0,viewport_coord[0][1]);
  
    viewport_coord[1][0] = my_max(point1[0],point2[0]);
    viewport_coord[1][0] = my_min(get_width(),viewport_coord[1][0]);
  
    viewport_coord[1][1] = my_max(point1[1],point2[1]);
    viewport_coord[1][1] = my_min(get_height(),viewport_coord[1][1]);
  }
  button_down[event->button - 1] = false;
  //invalidate();
  return true;
}

bool Viewer::on_motion_notify_event(GdkEventMotion* event) {
  //std::cerr << "Stub: Motion at " << event->x << ", " << event->y << std::endl;
  if(mode != VIEWPORT) {
    point2 = point1;
    point1[0] = event->x;
    point1[1] = event->y;
  }

  if(mode == MODEL_ROTATE) {
    if(button_down[0]) {
      m_model = m_model*rotation(point1[0]-point2[0],'x');
    }
    if(button_down[1]) {
      m_model = m_model*rotation(point1[0]-point2[0],'y');
    }
    if(button_down[2]) {
      m_model = m_model*rotation(point1[0]-point2[0],'z');
    }
    //std::cerr << rotation(point1[0]-point2[0],'x');
  }
  if(mode == MODEL_TRANSLATE) {
    Vector3D trans;
    if(button_down[0]) {
      trans[0] = (point1[0]-point2[0])/get_view_width()*10;
    }
    if(button_down[1]) {
      trans[1] = (point1[0]-point2[0])/get_view_width()*10;
    }
    if(button_down[2]) {
      trans[2] = (point1[0]-point2[0])/get_view_width()*10;
    }
    m_model=m_model*translation(trans);
    //std::cerr << translation(trans);
  }
  if(mode == MODEL_SCALE) {
    if(button_down[0]) {
      scale[0] = scale[0] * pow(1.01, point1[0]-point2[0]);
      if(scale[0] >= 5)    { scale[0] = 5; }
      if(scale[0] <= 0.05) { scale[0] = 0.05; }
    }
    if(button_down[1]) {
      scale[1] = scale[1] * pow(1.01, point1[0]-point2[0]);
      if(scale[1] >= 5)    { scale[1] = 5; }
      if(scale[1] <= 0.05) { scale[1] = 0.05; }
    }
    if(button_down[2]) {
      scale[2] = scale[2] * pow(1.01, point1[0]-point2[0]);
      if(scale[2] >= 5)    { scale[2] = 5; }
      if(scale[2] <= 0.05) { scale[2] = 0.05; }
    }
  }
  if(mode == VIEW_ROTATE) {
    if(button_down[0]) {
      m_view = rotation(point1[0]-point2[0],'x')*m_view;
    }
    if(button_down[1]) {
      m_view = rotation(point1[0]-point2[0],'y')*m_view;
    }
    if(button_down[2]) {
      m_view = rotation(point1[0]-point2[0],'z')*m_view;
    }
    //std::cerr << rotation(point1[0]-point2[0],'x');
  }
  if(mode == VIEW_TRANSLATE) {
    Vector3D trans;
    if(button_down[0]) {
      trans[0] = (point1[0]-point2[0])/get_view_width()*10;
    }
    if(button_down[1]) {
      trans[1] = (point1[0]-point2[0])/get_view_width()*10;
    }
    if(button_down[2]) {
      trans[2] = (point1[0]-point2[0])/get_view_width()*10;
    }
    m_view=translation(trans)*m_view;
    //std::cerr << translation(trans);
  }
  if(mode == VIEW_PERSPECTIVE) {
    if(button_down[0]) {
      fov += (point1[0]-point2[0]);
      if(fov < 5.0) {
        fov = 5.0;
      }
      if(fov > 160.0) {
        fov = 160.0;
      }
    }
    if(button_down[1]) {
      near_clip += 0.05*(point1[0]-point2[0]);
      if(near_clip < 0.1) {
        near_clip = 0.1;
      }
      if(near_clip < max_near_clip) {
        near_clip = max_near_clip;
      }
      near_clip = my_min(near_clip, far_clip);
    }
    if(button_down[2]) {
      far_clip += 0.05*(point1[0]-point2[0]);
      if(far_clip < 0.1) {
        far_clip = 0.1;
      }
      if(far_clip > max_far_clip) {
        far_clip = max_far_clip;
      }
      far_clip = my_max(near_clip, far_clip);
    }
  }
  //invalidate();
  return true;
}

void Viewer::set_mode(Mode aMode) {
  mode = aMode;
}
Point2D Viewer::get_view_center() {
  return Point2D((viewport_coord[0][0] + viewport_coord[1][0])/2,(viewport_coord[0][1] + viewport_coord[1][1])/2);
}

double Viewer::my_min(double a, double b) {
  if(a > b) {
    return b;
  }
  return a;
}

double Viewer::my_max(double a, double b) {
  if(a < b) {
    return b;
  }
  return a;
}

double Viewer::get_near() {
  return near_clip;
}

double Viewer::get_far() {
  return far_clip;
}
